unit dpobj;

// Copyright  2001 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking

//Various objects used by DiskPie

interface
uses Windows, sysutils, classes, forms;
type
  TSizList = class(TObject)
  // This object maintains a limited-size list of strings associated
  // with Int64 sizes. Items are ordered by size. If adding an item
  // makes the list too big, the smallest item is discarded.
  private
    fList : TStringList;
    fMax : Integer;
    function GetCount: Integer;
    function GetIName(Index: Integer): String;
    function GetISize(Index: Integer): Int64;
  protected
  public
    constructor Create(vMax : Integer);
    destructor Destroy; override;
    function Add(const S : String; Siz : Int64) : Boolean;
    procedure AddList(const Prefix : String; SL : TSizList);
    procedure Clear;
    property IName[Index : Integer] : String read GetIName;
    property ISize[Index : Integer] : Int64 read GetISize;
    property Count : Integer read GetCount;
  end;


  TUObj = class(TObject)
  // This simple-minded object simply serves to associate an Int64
  // with an object, so it can be added to a TStringList
  public
    Value : Int64;
    constructor Create(vValue : Int64);
    procedure AddValue(I : Int64);
  end;

  TDPNode = class(TObject)
  // DiskPie builds a skeletal folder tree, retaining the information
  // for any folders it has already examined. This object holds the
  // data for one folder
  private
    fSizList   : TSizList; // a list of the folder's largest files
    fName      : String;   // folder's name
    fSize      : Int64;    // size of files in and below folder
    fInSize    : Int64;    // size of files just IN folder
    fFileCount : Integer;  // count of files in and below folder
    fParent    : TDPNode;  // parent node
    fChildren  : TList;    // subfolders
    fPathname  : String;   // full path - calculated when requested
    function GetChild(Index: Integer): TDPNode;
    procedure SetChild(Index: Integer; const Value: TDPNode);
    function GetCount: Integer;
    function GetPathname: String;
    function CalcSize : Int64;
    procedure DoHog(TS : TSizList);
  protected
  public
    constructor Create(vParent : TDPNode; const vName : String);
    destructor Destroy; override;
    function ChildByName(const S : String) : TDPNode;
    procedure Expand;
    function GetSize: Int64;
    procedure NoSize;
    procedure LargestFiles(TS : TStringList);
    property AvailSize : Int64 read fSize;   // calculated size
    property InSize    : Int64 read fInSize; // size of files IN
    property Name      : String read fName;  // folder name
    property Pathname  : String read GetPathname; // pathname
    property Count     : Integer read GetCount;   // # of subfolders
    property FileCount : Integer read fFileCount; // # files in, below
    property Children[Index : Integer] : TDPNode read GetChild
      write SetChild;
  end;

  TDPTree = class(TObject)
  // This object represents a tree of TDPNodes
  private
    fRoot : TDPNode;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function NodeForPath(const S : String) : TDPNode;
    procedure Clear;
  published

  end;

  TDPTypeObj = class(TObject)
  // An object to total up the size for each file type
  private
    fList    : TStringList;
    fFilList : TSizList;
    fPath    : String;
    procedure AnalyzeOne(const S : String);
    procedure AddSize(const Ext : String; I : Int64);
    procedure AddFile(const Name : String; I : Int64);
  protected

  public
    constructor Create;
    destructor Destroy; override;
    procedure AnalyzePath(const vPath : String);
    procedure Clear;
    procedure LargestFiles(TS : TStringList);
    procedure ShowHogs(TS : TStringList; Num : Integer);
    class function AssnName(const Ext : String) : String;
  published

  end;


implementation
USES Registry, dpShare;


{ TSizList }

constructor TSizList.Create(vmax : Integer);
begin
  fList        := TStringList.Create;
  fList.Sorted := True;
  fMax         := vMax;
end;

destructor TSizList.Destroy;
begin
  fList.Free;
  inherited;
end;

function TSizList.Add(const S: String; Siz: Int64) : Boolean;
VAR NewS : String;
begin
  // Prefix the string with a 16-digit hex representation of
  // the size
  NewS   := IntToHex(Siz, 16) + S;
  Result := False;
  // If the list is full and it's less than the first item, skip it
  IF (fList.Count = fMax) AND (NewS < fList[0]) THEN
    Exit;
  Result := True;
  // Add to the sorted list
  fList.Add(NewS);
  // If the list is full, delete the first (smallest) item
  IF fList.Count > fMax THEN
    fList.Delete(0);
end;

procedure TSizList.AddList(const Prefix : String; SL: TSizList);
VAR N : Integer;
begin
  // Add another sorted list. Note that the IName and ISize properties
  // invert the order. Thus we're starting with the biggest item. The
  // first time one of them is too puny to add, we can quit.
  FOR N := 0 TO SL.Count-1 DO
    IF NOT Add(Prefix+SL.IName[N], SL.ISize[N]) THEN
      Exit;
end;

procedure TSizList.Clear;
begin
  fList.Clear;
end;

function TSizList.GetCount: Integer;
begin
  Result := fList.Count;
end;

function TSizList.GetIName(Index: Integer): String;
// The list is stored in smallest-first order, but the name and
// size values are returned in largest-first order
begin
  Result := fList[flist.Count-1-Index];
  System.Delete(Result, 1, 16);
end;

function TSizList.GetISize(Index: Integer): Int64;
// The list is stored in smallest-first order, but the name and
// size values are returned in largest-first order
begin
  Result := StrToInt64('$'+Copy(fList[flist.Count-1-Index], 1, 16));
end;

{ TUObj }

procedure TUObj.AddValue(I: Int64);
begin
  Value := Value + I;
end;

constructor TUObj.Create(vValue: Int64);
begin
  Inherited Create;
  Value := vValue;
end;

{ TDPNode }

constructor TDPNode.Create(vParent: TDPNode; const vName: String);
begin
  Inherited Create;
  fParent    := vParent;
  IF fParent <> nil THEN
    fParent.fChildren.Add(Self);
  fName      := vName;
  fSize      := -1;
  fInSize    := -1;
  fFileCount := -1;
  fPathname  := '';
  fChildren  := TList.Create;
  fSizList   := TSizList.Create(uttermax);
end;

destructor TDPNode.Destroy;
VAR N : Integer;
begin
  FOR N := Count-1 DOWNTO 0 DO
    Children[N].Free;
  fChildren.Free;
  inherited;
end;

function TDPNode.ChildByName(const S: String): TDPNode;
// Return the child node matching the name, if any
VAR N : Integer;
begin
  FOR N := 0 TO Count-1 DO
    begin
      IF AnsiCompareText(Children[N].Name, S) = 0 THEN
        begin
          Result := Children[N];
          Exit;
        end;
    end;
  Result := nil;
end;

procedure TDPNode.Expand;
// Check the disk and add the folder's children. Do so recursively
// if asked.
VAR
  FH : THandle;
  FD : TWin32FindData;
  N  : Integer;
begin
  FH := FindFirstFile(PChar(Pathname+'\*.*'), FD);
  IF FH = INVALID_HANDLE_VALUE THEN Exit;
  try
    REPEAT
      Application.ProcessMessages;
      IF Halting THEN Exit;
      IF FD.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY = 0 THEN
        Continue;
      IF AnsiCompareText(FD.cFileName, '.') = 0 THEN Continue;
      IF AnsiCompareText(FD.cFileName, '..') = 0 THEN Continue;
      IF ChildByName(FD.cFilename) = nil THEN
        TDPNode.Create(Self, FD.cFilename);
    UNTIL NOT FindNextFile(FH, FD);
  finally
    Windows.FindClose(FH);
  end;
  FOR N := 0 TO COunt-1 DO
    Children[N].Expand;
end;

function TDPNode.GetSize: Int64;
// If the total size has not been calculated, calculate it now.
// Either way, return the value.
begin
  IF fSize < 0 THEN
    fSize := CalcSize;
  Result :=fSize;
end;

procedure TDPNode.NoSize;
// Clear the total size information for this node and children
VAR N : Integer;
begin
  FOR N := 0 TO Count-1 DO
    Children[N].NoSize;
  fSizList.Clear;
  fSize := -1;
end;

procedure TDPNode.LargestFiles(TS: TStringList);
// Figure out the largest files in and below this node. Place that
// information in the passed TStringList
VAR
  TmpS : TSizList;
  N    : Integer;
begin
  TS.Clear;
  TmpS := TSizList.Create(UtterMax);
  try
    DoHog(TmpS);
    FOR N := 0 TO TmpS.Count-1 DO
      TS.AddObject(TmpS.IName[N],
        TUobj.Create(TmpS.ISize[N]));
  finally
    TmpS.Free;
  end;
end;

function TDPNode.GetChild(Index: Integer): TDPNode;
begin
  IF Index >= fChildren.Count THEN
    Raise(Exception.Create('Index out of range in TDPNode.GetChild'));
  Result := TDPNode(fChildren[Index]);
end;

procedure TDPNode.SetChild(Index: Integer; const Value: TDPNode);
begin
  IF Index >= fChildren.Count THEN
    Raise(Exception.Create('Index out of range in TDPNode.SetChild'));
  fChildren[Index] := Value;
end;

function TDPNode.GetCount: Integer;
begin
  Result := fChildren.Count;
end;

function TDPNode.GetPathname: String;
begin
  IF fPathname <> '' THEN Result := fPathname
  ELSE
    begin
      IF (fParent.fParent = nil) THEN Result := fName
      ELSE Result := fParent.GetPathname + '\' + fName;
      fPathname := Result;
    end;
end;

function TDPNode.CalcSize: Int64;
// Calculate the size of files in this folder, the size of files
// in and below this folder, and the 20 largest files
VAR
  FH  : THandle;
  FD  : TWin32FindData;
  Siz : ULARGE_INTEGER;
  N   : Integer;
begin
  fInSize    := 0;
  fFileCount := 0;
  FH := FindFirstFile(PChar(Pathname+'\*.*'), FD);
  IF FH = INVALID_HANDLE_VALUE THEN
    begin
      Result := -1; //NOTE: Previously left it at 0
      Exit;
    end;
  try
    REPEAT
      Application.ProcessMessages;
      IF Halting THEN Exit;
      IF FD.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY > 0 THEN
        Continue;
      Siz.LowPart  := FD.nFileSizeLow;
      Siz.HighPart := FD.nFileSizeHigh;
      fInSize      := fInsize + Siz.QuadPart;
      fSizList.Add(FD.cFilename, Siz.QuadPart);
      Inc(fFileCount);
    UNTIL NOT FindNextFile(FH, FD);
  finally
    Windows.FindClose(FH);
  end;
  Result := fInSize;
  FOR N := 0 TO Count-1 DO
    begin
      Result := Result + Children[N].GetSize;
      Inc(fFileCount, Children[N].FileCount);
    end;
end;

procedure TDPNode.DoHog(TS: TSizList);
// Add this folder's largest files to the TSizList. Then add its
// children's largest files. Only the very largest will stay
// in the TSizList.
VAR N : Integer;
begin
  TS.AddList(Self.fPathname+'\', fSizList);
  FOR N := 0 TO COunt-1 DO
    Children[N].DoHog(TS);
end;

{ TDPTree }

constructor TDPTree.Create;
begin
  Inherited Create;
  fRoot := TDPNode.Create(nil, '');
end;

destructor TDPTree.Destroy;
begin
  fRoot.Free;
  inherited;
end;

procedure TDPTree.Clear;
begin
  fRoot.Free;
  fRoot := TDPNode.Create(nil, '');
end;

function TDPTree.NodeForPath(const S: String): TDPNode;
// Return the node corresponding to this path, if it exists. If not
// create it (and all necessary parents)
VAR
  sName  : String;
  sPath  : String;
  Parent : TDPNode;
begin
  IF S = '' THEN Result := fRoot
  ELSE
    begin
      sPath := S;
      // Knock off any final backslash
      IF (sPath <> '') AND (sPath[Length(sPath)] = '\') THEN
        SetLength(sPath, Length(sPath)-1);
      sName := ExtractFileName(sPath);
      sPath := ExtractFileDir(sPath);
      // If this is a root directory, set its parent to the root node
      IF sName = '' THEN
        begin
          Parent := fRoot;
          sName := sPath;
        end
      // Otherwise, call NodeForPath recursively to get the parent
      ELSE Parent := NodeForPath(sPath);
      // Return parent's matching child node
      Result := Parent.ChildByName(sName);
      // If no matching node, create it
      IF Result = nil THEN
        Result := TDPNode.Create(Parent, sName);
    end;
end;

{ TDPTypeObj }

constructor TDPTypeObj.Create;
begin
  Inherited Create;
  fList := TStringList.Create;
  fFilList := TSizList.Create(uttermax);
  fList.Sorted := True;
end;

destructor TDPTypeObj.Destroy;
begin
  Clear;
  fList.Free;
  inherited;
end;

procedure TDPTypeObj.AnalyzePath(const vPath: String);
VAR TPath : String;
begin
  IF vPath = '' THEN Exit;
  IF vPath = '\' THEN Exit;
  tPath := vPath;
  IF tPath[Length(tPath)] = '\' THEN
    SetLength(tPath, Pred(Length(tPath)));
  IF tPath = fPath THEN Exit;
  IF NOT DirectoryExists(tPath) THEN Exit;
  fPath := tPath;
  Clear;
  AnalyzeOne(tPath);
end;

procedure TDPTypeObj.Clear;
VAR N : Integer;
begin
  FOR N := 0 TO fList.Count-1 DO
    TUObj(fList.Objects[N]).Free;
  fList.Clear;
  fFilList.Clear;
  fFilList.Clear;
end;

procedure TDPTypeObj.LargestFiles(TS: TStringList);
VAR N : Integer;
begin
  FOR N := 0 TO fFilList.Count-1 DO
    TS.AddObject(fFilList.IName[N],
      TUobj.Create(fFilList.ISize[N]));
end;

  function Cust(List: TStringList; Index1, Index2: Integer): Integer;
  VAR I1, I2 : Int64;
  begin
    I1 := TUobj(List.Objects[Index1]).Value;
    I2 := TUobj(List.Objects[Index2]).Value;
    IF I1 < I2 THEN Result := -1
    ELSE IF I1 > I2 THEN Result := 1
    ELSE Result := 0;
    // We want REVERSE size order
    Result := -Result;
  end;

procedure TDPTypeObj.ShowHogs(TS: TStringList; Num : Integer);
//Assumes TS is empty
VAR
  Other : Int64;
  N     : Integer;
begin
  fList.Sorted := False;
  try
    fList.CustomSort(Cust);
    IF fList.Count <= Num THEN
      begin
        FOR N := 0 TO fList.Count-1 DO
          IF fList[N] = '' THEN
            TS.AddObject('(none)', TUobj.Create(TUobj(fList.Objects[N]).Value))
          ELSE
            TS.AddObject(fList[N], TUobj.Create(TUobj(fList.Objects[N]).Value));
      end
    ELSE
      begin
        Other := 0;
        FOR N := fList.Count-1 DOWNTO Num-1 DO
          Other := Other + TUObj(fList.Objects[N]).Value;
        FOR N := 0 TO Num-2 DO
          IF fList[N] = '' THEN
            TS.AddObject('(none)', TUobj.Create(TUobj(fList.Objects[N]).Value))
          ELSE
            TS.AddObject(fList[N], TUobj.Create(TUobj(fList.Objects[N]).Value));
        TS.AddObject('"Other"', TUobj.Create(Other));
      end;
  finally
    fList.Sorted := True;
  end;
end;

//!!
procedure TDPTypeObj.AddFile(const Name: String; I: Int64);
begin
  fFilList.Add(Name, I);
end;

procedure TDPTypeObj.AddSize(const Ext: String; I: Int64);
VAR Idx : Integer;
begin
  Idx := fList.IndexOf(Ext);
  IF Idx < 0 THEN
    fList.AddObject(Ext, TUobj.Create(I))
  ELSE
    TUobj(fList.Objects[Idx]).AddValue(I);
end;

procedure TDPTypeObj.AnalyzeOne(const S: String);
VAR
  FH : THandle;
  FD : TWin32FindData;
  Siz : ULARGE_INTEGER;
begin
  FH := FindFirstFile(PChar(S+'\*.*'), FD);
  IF FH = INVALID_HANDLE_VALUE THEN Exit;
  try
    REPEAT
      Application.ProcessMessages;
      IF Halting THEN Exit;
      IF FD.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY > 0 THEN
        begin
          IF AnsiCompareText(FD.cFileName, '.') = 0 THEN Continue;
          IF AnsiCompareText(FD.cFileName, '..') = 0 THEN Continue;
          AnalyzeOne(S+'\'+FD.cFilename);
        end
      ELSE
        begin
          Siz.LowPart := FD.nFileSizeLow;
          Siz.HighPart := FD.nFileSizeHigh;
          AddSize(UpperCase(ExtractFileExt(FD.cFilename)),
            Siz.QuadPart);
          AddFile(S+'\'+FD.cFilename, Siz.QuadPart);
        end;
    UNTIL NOT FindNextFile(FH, FD);
  finally
    Windows.FindClose(FH);
  end;
end;

class function TDPTypeObj.AssnName(const Ext : String) : String;
VAR disp : String;
begin
  Result := Ext;
  IF (Ext = '(none)') OR (Ext = '"Other"') THEN Exit;
  WITH TRegistry.Create DO
  try
    RootKey := HKEY_CLASSES_ROOT;
    IF OpenKeyReadOnly(Ext) THEN
      IF ValueExists('') THEN
        IF GetDataType('') = rdString THEN
          begin
            disp := ReadString('');
            IF OpenKeyReadOnly('\'+disp) THEN
              IF ValueExists('') THEN
                IF GetDataType('') = rdString THEN
                  Result := Result + ' - '+ReadString('');
          end;
  finally
    Free;
  end;
end;

initialization
  Halting := False;
end.
